"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
from fastapi import HTTPException

# // Main
class PTSException(Exception):
    """
    Base class for all exceptions in PythonToSW.
    """
    
class PTSCallException(PTSException):
    """
    Raised when something goes wrong with addon calls.
    """

class PTSCallbackException(PTSException):
    """
    Raised when something goes wrong with addon callbacks.
    """
    
class PTSLifecycleException(PTSException):
    """
    Raised when something goes wrong with addon lifecycle.
    """
    
class PTSConfigException(PTSException):
    """
    Raised when something goes wrong with addon configuration.
    """
    
class PTSHTTPException(PTSException, HTTPException):
    """
    Exception class for HTTP errors in PythonToSW.
    Inherits from both `PTSException` and FastAPI's `HTTPException`.
    """

    def __init__(self, status_code: int, type: str, detail: str):
        """
        Initializes a new instance of the `PTSHTTPException` class.
        
        Args:
            status_code (int): The HTTP status code for the exception.
            type (str): The type of the exception.
            detail (str): The detail message for the exception.
        """

        PTSException.__init__(self, detail)
        HTTPException.__init__(self, status_code = status_code, detail = f"{type}: {detail}")


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
from enum import Enum

# // Main
__all__ = [
    "CallEnum",
    "CallbackEnum"
]

class CallbackEnum(Enum):
    """
    An enum for all the available callbacks in Stormworks.
    """
    
    ON_CLEAR_OIL_SPILL = "onClearOilSpill"
    ON_CREATE = "onCreate"
    ON_DESTROY = "onDestroy"
    ON_CUSTOM_COMMAND = "onCustomCommand"
    ON_CHAT_MESSAGE = "onChatMessage"
    ON_PLAYER_JOIN = "onPlayerJoin"
    ON_PLAYER_SIT = "onPlayerSit"
    ON_PLAYER_UNSIT = "onPlayerUnsit"
    ON_CHARACTER_SIT = "onCharacterSit"
    ON_CHARACTER_UNSIT = "onCharacterUnsit"
    ON_CHARACTER_PICKUP = "onCharacterPickup"
    ON_CREATURE_SIT = "onCreatureSit"
    ON_CREATURE_UNSIT = "onCreatureUnsit"
    ON_CREATURE_PICKUP = "onCreaturePickup"
    ON_EQUIPMENT_PICKUP = "onEquipmentPickup"
    ON_EQUIPMENT_DROP = "onEquipmentDrop"
    ON_PLAYER_RESPAWN = "onPlayerRespawn"
    ON_PLAYER_LEAVE = "onPlayerLeave"
    ON_TOGGLE_MAP = "onToggleMap"
    ON_PLAYER_DIE = "onPlayerDie"
    ON_VEHICLE_SPAWN = "onVehicleSpawn"
    ON_GROUP_SPAWN = "onGroupSpawn"
    ON_VEHICLE_DESPAWN = "onVehicleDespawn"
    ON_VEHICLE_LOAD = "onVehicleLoad"
    ON_VEHICLE_UNLOAD = "onVehicleUnload"
    ON_VEHICLE_TELEPORT = "onVehicleTeleport"
    ON_OBJECT_LOAD = "onObjectLoad"
    ON_OBJECT_UNLOAD = "onObjectUnload"
    ON_BUTTON_PRESS = "onButtonPress"
    ON_SPAWN_ADDON_COMPONENT = "onSpawnAddonComponent"
    ON_VEHICLE_DAMAGED = "onVehicleDamaged"
    ON_HTTP_REPLY = "httpReply"
    ON_FIRE_EXTINGUISHED = "onFireExtinguished"
    ON_FOREST_FIRE_SPAWNED = "onForestFireSpawned"
    ON_FOREST_FIRE_EXTINGUISHED = "onForestFireExtinguished"
    ON_TORNADO = "onTornado"
    ON_METEOR = "onMeteor"
    ON_TSUNAMI = "onTsunami"
    ON_WHIRLPOOL = "onWhirlpool"
    ON_VOLCANO = "onVolcano"
    ON_OIL_SPILL = "onOilSpill"

class CallEnum(Enum):
    """
    An enum for all the available server functions that can be called in Stormworks.
    """
    
    GET_ADDON_INDEX = "getAddonIndex"
    GET_PLAYLIST_INDEX_CURRENT = "getPlaylistIndexCurrent"
    GET_PLAYLIST_INDEX_BY_NAME = "getPlaylistIndexByName"
    GET_LOCATION_INDEX = "getLocationIndex"
    SPAWN_THIS_ADDON_LOCATION = "spawnThisAddonLocation"
    SPAWN_THIS_PLAYLIST_MISSION_LOCATION = "spawnThisPlaylistMissionLocation"
    SPAWN_NAMED_ADDON_LOCATION = "spawnNamedAddonLocation"
    SPAWN_ADDON_LOCATION = "spawnAddonLocation"
    SPAWN_MISSION_LOCATION = "spawnMissionLocation"
    GET_ADDON_PATH = "getAddonPath"
    GET_PLAYLIST_PATH = "getPlaylistPath"
    GET_ZONES = "getZones"
    IS_IN_ZONE = "isInZone"
    GET_ADDON_COUNT = "getAddonCount"
    GET_PLAYLIST_COUNT = "getPlaylistCount"
    GET_ADDON_DATA = "getAddonData"
    GET_PLAYLIST_DATA = "getPlaylistData"
    GET_LOCATION_DATA = "getLocationData"
    GET_LOCATION_COMPONENT_DATA = "getLocationComponentData"
    SPAWN_ADDON_COMPONENT = "spawnAddonComponent"
    SPAWN_MISSION_COMPONENT = "spawnMissionComponent"
    ANNOUNCE = "announce"
    COMMAND = "command"
    NOTIFY = "notify"
    GET_MAP_ID = "getMapID"
    REMOVE_MAP_ID = "removeMapID"
    ADD_MAP_OBJECT = "addMapObject"
    REMOVE_MAP_OBJECT = "removeMapObject"
    ADD_MAP_LABEL = "addMapLabel"
    REMOVE_MAP_LABEL = "removeMapLabel"
    ADD_MAP_LINE = "addMapLine"
    REMOVE_MAP_LINE = "removeMapLine"
    SET_POPUP = "setPopup"
    SET_POPUP_SCREEN = "setPopupScreen"
    REMOVE_POPUP = "removePopup"
    GET_PLAYERS = "getPlayers"
    GET_PLAYER_NAME = "getPlayerName"
    GET_PLAYER_POS = "getPlayerPos"
    SET_PLAYER_POS = "setPlayerPos"
    GET_PLAYER_LOOK_DIRECTION = "getPlayerLookDirection"
    GET_PLAYER_CHARACTER_ID = "getPlayerCharacterID"
    SPAWN_OBJECT = "spawnObject"
    SPAWN_FIRE = "spawnFire"
    SPAWN_CHARACTER = "spawnCharacter"
    SPAWN_ANIMAL = "spawnAnimal"
    SPAWN_CREATURE = "spawnCreature"
    SET_CREATURE_MOVE_TARGET = "setCreatureMoveTarget"
    SPAWN_EQUIPMENT = "spawnEquipment"
    DESPAWN_OBJECT = "despawnObject"
    GET_OBJECT_POS = "getObjectPos"
    GET_OBJECT_SIMULATING = "getObjectSimulating"
    SET_OBJECT_POS = "setObjectPos"
    SET_FIRE_DATA = "setFireData"
    GET_FIRE_DATA = "getFireData"
    KILL_CHARACTER = "killCharacter"
    REVIVE_CHARACTER = "reviveCharacter"
    SET_SEATED = "setSeated"
    SET_CHARACTER_SEATED = "setCharacterSeated"
    SET_CREATURE_SEATED = "setCreatureSeated"
    GET_OBJECT_DATA = "getObjectData"
    GET_CHARACTER_DATA = "getCharacterData"
    GET_CHARACTER_VEHICLE = "getCharacterVehicle"
    SET_CHARACTER_DATA = "setCharacterData"
    SET_CHARACTER_TOOLTIP = "setCharacterTooltip"
    SET_CREATURE_TOOLTIP = "setCreatureTooltip"
    SET_CHARACTER_ITEM = "setCharacterItem"
    GET_CHARACTER_ITEM = "getCharacterItem"
    SPAWN_ADDON_VEHICLE = "spawnAddonVehicle"
    SPAWN_VEHICLE = "spawnVehicle"
    SPAWN_VEHICLE_SAVE_FILE = "spawnVehicleSavefile"
    DESPAWN_VEHICLE = "despawnVehicle"
    IS_LOCATION_CLEAR = "isLocationClear"
    GET_VEHICLE_POS = "getVehiclePos"
    SET_VEHICLE_POS = "setVehiclePos"
    SET_VEHICLE_POS_SAFE = "setVehiclePosSafe"
    RESET_VEHICLE_STATE = "resetVehicleState"
    CLEAN_VEHICLES = "cleanVehicles"
    CLEAR_RADIATION = "clearRadiation"
    GET_VEHICLE_DATA = "getVehicleData"
    GET_VEHICLES_BY_NAME = "getVehiclesByName"
    GET_VEHICLE_COMPONENTS = "getVehicleComponents"
    GET_VEHICLE_GROUP = "getVehicleGroup"
    DESPAWN_VEHICLE_GROUP = "despawnVehicleGroup"
    SET_GROUP_POS = "setGroupPos"
    MOVE_GROUP = "moveGroup"
    MOVE_GROUP_SAFE = "moveGroupSafe"
    MOVE_VEHICLE = "moveVehicle"
    MOVE_VEHICLE_SAFE = "moveVehicleSafe"
    SET_GROUP_POS_SAFE = "setGroupPosSafe"
    GET_ASTROPOS = "getAstroPos"
    GET_VEHICLE_TANK = "getVehicleTank"
    GET_VEHICLE_SIGN = "getVehicleSign"
    GET_VEHICLE_SEAT = "getVehicleSeat"
    GET_VEHICLE_BUTTON = "getVehicleButton"
    GET_VEHICLE_DIAL = "getVehicleDial"
    GET_VEHICLE_BATTERY = "getVehicleBattery"
    GET_VEHICLE_HOPPER = "getVehicleHopper"
    GET_VEHICLE_WEAPON = "getVehicleWeapon"
    GET_VEHICLE_ROPE_HOOK = "getVehicleRopeHook"
    SET_VEHICLE_SEAT = "setVehicleSeat"
    PRESS_VEHICLE_BUTTON = "pressVehicleButton"
    SET_VEHICLE_KEYPAD = "setVehicleKeypad"
    SET_VEHICLE_TANK = "setVehicleTank"
    SET_VEHICLE_BATTERY = "setVehicleBattery"
    SET_VEHICLE_HOPPER = "setVehicleHopper"
    SET_VEHICLE_WEAPON = "setVehicleWeapon"
    SPAWN_VEHICLE_ROPE = "spawnVehicleRope"
    GET_VEHICLE_FIRE_COUNT = "getVehicleFireCount"
    SET_VEHICLE_TOOLTIP = "setVehicleTooltip"
    ADD_DAMAGE = "addDamage"
    GET_VEHICLES_SIMULATING = "getVehicleSimulating"
    GET_VEHICLE_LOCAL = "getVehicleLocal"
    SET_VEHICLE_TRANSPONDER = "setVehicleTransponder"
    SET_VEHICLE_EDITABLE = "setVehicleEditable"
    SET_VEHICLE_INVULNERABLE = "setVehicleInvulnerable"
    SET_VEHICLE_SHOW_ON_MAP = "setVehicleShowOnMap"
    SET_AI_STATE = "setAIState"
    SET_AI_TARGET = "setAITarget"
    GET_AI_TARGET = "getAITarget"
    SET_AI_TARGET_CHARACTER = "setAITargetCharacter"
    SET_AI_TARGET_VEHICLE = "setAITargetVehicle"
    SET_AI_CHARACTER_TEAM = "setAICharacterTeam"
    SET_AI_VEHICLE_TEAM = "setAIVehicleTeam"
    SET_AI_CHARACTER_TARGET_TEAM = "setAICharacterTargetTeam"
    GET_FISH_DATA = "getFishData"
    GET_FISH_HOTSPOTS = "getFishHotspots"
    SPAWN_TSUNAMI = "spawnTsunami"
    SPAWN_WHIRLPOOL = "spawnWhirlpool"
    CANCEL_GERSTNER = "cancelGerstner"
    SPAWN_TORNADO = "spawnTornado"
    SPAWN_METEOR = "spawnMeteor"
    SPAWN_METEOR_SHOWER = "spawnMeteorShower"
    SPAWN_VOLCANO = "spawnVolcano"
    GET_VOLCANOS = "getVolcanos"
    GET_OIL_SPILL = "getOilSpill"
    SET_OIL_SPILL = "setOilSpill"
    CLEAR_OIL_SPILL = "clearOilSpill"
    SPAWN_EXPLOSION = "spawnExplosion"
    SET_GAME_SETTING = "setGameSetting"
    GET_GAME_SETTINGS = "getGameSettings"
    SET_CURRENCY = "setCurrency"
    GET_CURRENCY = "getCurrency"
    GET_RESEARCH_POINTS = "getResearchPoints"
    GET_DATE_VALUE = "getDateValue"
    GET_DATE = "getDate"
    GET_TIME = "getTime"
    GET_WEATHER = "getWeather"
    SET_WEATHER = "setWeather"
    SET_AUDIO_MOOD = "setAudioMood"
    GET_OCEAN_TRANSFORM = "getOceanTransform"
    GET_OCEAN_FLOOR = "getOceanFloor"
    GET_TILE_TRANSFORM = "getTileTransform"
    GET_TILE = "getTile"
    GET_START_TILE = "getStartTile"
    GET_START_ISLAND = "getStartIsland"
    GET_TILE_PURCHASE = "getTilePurchased"
    GET_TILE_INVENTORY = "getTileInventory"
    SET_TILE_INVENTORY = "setTileInventory"
    IS_IN_TRANSFORM_AREA = "isInTransformArea"
    PATH_FIND = "pathfind"
    PATH_FIND_OCEAN = "pathfindOcean"
    GET_OIL_DEPOSITS = "getOilDeposits"
    BAN_PLAYER = "banPlayer"
    KICK_PLAYER = "kickPlayer"
    ADD_ADMIN = "addAdmin"
    REMOVE_ADMIN = "removeAdmin"
    ADD_AUTH = "addAuth"
    REMOVE_AUTH = "removeAuth"
    SAVE = "save"
    GET_TIME_MILLISEC = "getTimeMillisec"
    GET_TUTORIAL = "getTutorial"
    SET_TUTORIAL = "setTutorial"
    GET_VIDEO_TUTORIAL = "getVideoTutorial"
    IS_DEV = "isDev"
    DLC_WEAPONS = "dlcWeapons"
    DLC_ARID = "dlcArid"
    DLC_SPACE = "dlcSpace"
    GET_SEASONAL_EVENT = "getSeasonalEvent"


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
import os
PACKAGE_PATH = os.path.dirname(os.path.abspath(__file__))

from . import log
from .log import logger

from .libs import (
    io,
    xml,
    http
)

from .libs.persistence import Persistence
from .libs.event import Event

from . import exceptions

from .values import *
from .enums import *
from .models import *

from .addon import *

# // Main
from logging import INFO as _INFO
log.install(_INFO)


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
from pydantic import (
    BaseModel,
    Field,
    field_serializer,
    SerializationInfo,
    ConfigDict
)

from typing import (
    Union,
    Any
)

from concurrent.futures import Future

from . import CallEnum
from . import BaseValue

# // Main
__all__ = [
    "Call",
    "Token"
]

class Token(BaseModel):
    """
    Represents a token for an addon.
    """
    
    token: str
    set_at: float

class Call(BaseModel):
    """
    Represents a call to a function in the addon.
    """
    
    model_config = ConfigDict(
        arbitrary_types_allowed = True
    )

    id: str
    path: str
    arguments: list[Union[Any, BaseValue]]
    future: Future = Field(default_factory = Future, exclude = True)
    
    @field_serializer("arguments")
    def serialize_arguments(self, arguments: Union[Any, BaseValue], _info: SerializationInfo):
        """
        Converts custom values to their Stormworks representation during serialization.
        
        Args:
            arguments (Union[Any, BaseValue]): The list of arguments to validate.
            _info (SerializationInfo): Serialization information.
        
        Returns:
            list: The validated list of arguments.
        """
        
        for index, argument in enumerate(arguments):
            if BaseValue.is_value(argument):
                arguments[index] = argument.build()
                
        return arguments


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
import pytest

from string import (
    ascii_letters,
    punctuation,
    digits,
    whitespace
)

from urllib.parse import quote

from PythonToSW import http

# // Main
@pytest.fixture(scope = "module") # change to "function" if randomness is added to the code below or something
def full() -> str:
    """
    Creates a string with all ASCII letters, punctuation, digits, and whitespace
    
    Returns:
        str: The full string containing all characters
    """
    
    return ascii_letters + punctuation + digits + whitespace

def test_encode(full: str):
    """
    Tests if URL encoding a string works
    
    Args:
        full (str): The string to encode
    """    
    
    assert http.url_encode(full) == quote(full), "URL encoding did not return expected result"
    
def test_decode(full: str):
    """
    Tests if URL decoding a string works
    
    Args:
        full (str): The string to decode
    """
    
    assert http.url_decode(http.url_encode(full)) == full, "URL decoding did not return expected result"


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
import pytest

from PythonToSW import Event

# // Main
@pytest.fixture(scope = "function")
def event() -> Event:
    """
    Creates an Event instance
    
    Returns:
        Event: The Event instance
    """    
    
    return Event()

def test_subscribe(event: Event):
    """
    Tests if subscribing to an event works
    
    Args:
        event (Event): The Event instance
    """    
    
    callback_1 = lambda: None
    callback_2 = lambda: None
    
    event.subscribe(callback_1)
    event += callback_2
    
    assert callback_1 in event._callbacks, "Callback 1 should be in event callbacks after subscribe (error with `.subscribe()` method)"
    assert callback_2 in event._callbacks, "Callback 2 should be in event callbacks after subscribe (error with `+=` operator)"
    
def test_unsubscribe(event: Event):
    """
    Tests if unsubscribing from an event works
    
    Args:
        event (Event): The Event instance
    """    
    
    callback_1 = lambda: None
    callback_2 = lambda: None

    event.subscribe(callback_1)
    event += callback_2
    
    event.unsubscribe(callback_1)
    event -= callback_2
    
    assert callback_1 not in event._callbacks, "Callback 1 should not be in event callbacks after unsubscribe (error with `.unsubscribe()` method)"
    assert callback_2 not in event._callbacks, "Callback 2 should not be in event callbacks after unsubscribe (error with `-=` operator)"
    
def test_fire(event: Event):
    """
    Tests if firing an event works
    
    Args:
        event (Event): The Event instance
    """    
    
    count = 0
    
    def increment_count():
        nonlocal count
        count += 1
    
    event.subscribe(increment_count)
    event += increment_count

    event.fire()
    event()
    
    assert count == 4, "Event should have been fired twice (firing two callbacks), count is not 4"
    
@pytest.mark.asyncio
async def test_fire_async(event: Event):
    """
    Tests if firing an async event works
    
    Args:
        event (Event): The Event instance
    """    
    
    count = 0
    
    async def async_callback():
        nonlocal count
        count += 1
    
    event.subscribe(async_callback)

    await event.fire_async()
    
    assert count == 1, "Event should have been fired once, count is not 1"


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
from __future__ import annotations

import math
from . import BaseValue

# // Main
class Matrix(BaseValue):
    """
    A class representing a matrix in Stormworks.
    """
    
    def __init__(self, x: float|int, y: float|int, z: float|int):
        """
        Initializes a new instance of the `Matrix` class.

        Args:
            x (float | int): The x-coordinate of the matrix.
            y (float | int): The y-coordinate of the matrix.
            z (float | int): The z-coordinate of the matrix.
        """        

        self.x = x
        self.y = y
        self.z = z
        
    def build(self) -> list:
        """
        Builds the matrix into a format suitable for Stormworks.

        Returns:
            list: The built matrix.
        """

        return [
            1, 0, 0, 0,
            0, 0, 0, 0,
            0, 0, 1, 0,
            self.x, self.y, self.z, 1
        ]
        
    @classmethod
    def rebuild(cls, value: list) -> Matrix:
        """
        Rebuilds a matrix from a Stormworks value.

        Args:
            value (list): The Stormworks value to rebuild from.
            
        Raises:
            ValueError: If the value does not have the expected length of 16 elements.

        Returns:
            Matrix: The rebuilt matrix.
        """
        
        if not isinstance(value, list):
            raise ValueError("Invalid matrix value type. Expected a list.")

        if len(value) != 16:
            raise ValueError("Invalid matrix value length. Expected 16 elements.")

        return cls(
            x = value[12],
            y = value[13],
            z = value[14]
        )

    def distance(self, other: Matrix) -> float:
        """
        Returns the distance between this matrix and another matrix.

        Returns:
            float: The distance between the two matrices.
        """

        return math.sqrt(
            (self.x - other.x) ** 2 +
            (self.y - other.y) ** 2 +
            (self.z - other.z) ** 2
        )


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
from __future__ import annotations
from typing import Any
from abc import abstractmethod, ABCMeta

# // Main
class BaseValue(metaclass = ABCMeta):
    """
    Base class for all values in PythonToSW.
    """
    
    @staticmethod
    def is_value(instance: BaseValue):
        """
        Checks if the instance is a value.
        
        Args:
            instance (BaseValue): The instance to check.
        
        Returns:
            bool: True if the instance is a value, False otherwise.
        """
        
        return isinstance(instance, BaseValue)

    @abstractmethod
    def build(self):
        """
        Builds the value into a format suitable for Stormworks.
        
        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        
        Returns:
            The built value.
        """
        
        raise NotImplementedError("The `build` method must be implemented by subclasses.")
    
    @abstractmethod
    def rebuild(cls, value: Any) -> BaseValue:
        """
        Creates a value from the serialized variant (`.build()` return value).
        
        Args:
            value (Any): The value to convert.
            
        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        
        Returns:
            BaseValue: The converted value.
        """
        
        raise NotImplementedError("The `rebuild` method must be implemented by subclasses.")


"""
----------------------------------------------
PythonToSW: A Python package that allows you to make Stormworks addons with Python.
https://github.com/Cuh4/PythonToSW
----------------------------------------------

Copyright (C) 2025 Cuh4

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# // Imports
import os
from setuptools import setup, find_packages

# // Main
with open(os.path.join(os.path.dirname(__file__), "VERSION"), encoding = "utf-8") as file:
    version = file.read()
    
with open(os.path.join(os.path.dirname(__file__), "README.md"), encoding = "utf-8") as file:
    long_description = file.read()
    
with open(os.path.join(os.path.dirname(__file__), "requirements.txt"), encoding = "utf-8") as file:
    requirements = file.read().splitlines()

setup(
    name = "PythonToSW",
    version = version,
    author = "Cuh4",
    description = "A package that allows you to create addons in Stormworks with Python, handled through HTTP.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    packages = find_packages(where = "src"),
    license = "Apache License 2.0",
    url = "https://github.com/Cuh4/PythonToSW",
    
    classifiers = [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    
    project_urls = {
        "Source code": "https://github.com/Cuh4/PythonToSW"
    },

    install_requires = requirements,

    python_requires = ">=3.12",
    include_package_data = True,
    package_dir = {"": "src"}
)